# -*- coding: utf-8 -*-
import subprocess
import time
from pathlib import Path

import click

from ros_d2.export import export


@click.group()
def cli():
    pass


@cli.command("export", help="Export ROS 2 node graph to D2")
@click.option(
    "--verbose",
    is_flag=True,
    default=False,
    help="Make the diagram more verbose with topic types and more.",
)
@click.argument("output", type=str)
def export_command(output: str, verbose: bool):
    output_file = Path(output)

    # If the output_file is not a .d2 file, add the extension.
    if output_file.suffix != ".d2":
        output_file = output_file.with_suffix(".d2")

    export(output_file, verbose)


@cli.command(
    "render", help="Find all .d2 files in the current directory and render them"
)
@click.option(
    "--layout",
    default="dagre",
    help="The d2 layout type",
)
def render_command(layout: str):
    d2_files = Path(".").glob("**/*.d2")
    d2_files_list = list(d2_files)
    click.echo(f"D2 files found: {len(d2_files_list)}")
    for d2_file in d2_files_list:
        # Execute d2 render command
        command = ["d2", "--layout", layout, str(d2_file)]
        click.echo(f"Running: {' '.join(command)}")
        subprocess.run(command)


@cli.command(
    "export-launch-files",
    help="Finds all your .launch.py files, runs them and exports the resulting graph",
)
@click.option(
    "--wait",
    default=2,
    help="How long to wait for the launch file to start before exporting the graph",
)
def export_launch_files_command(wait: int):
    launch_files = Path(".").glob("**/*.launch.py")
    launch_files_list = list(launch_files)
    click.echo(f"ROS launch files found: {len(launch_files_list)}")
    for launch_file in launch_files_list:
        # Execute d2 render command
        launch_file_name = launch_file.stem
        # replace .launch.py with .d2
        output_file = launch_file.with_name(launch_file_name + ".d2")
        command_launch = [
            "ros2",
            "launch",
            str(launch_file),
        ]
        click.echo(f"Running: {' '.join(command_launch)}")
        # Run this command in a subprocess so we can end the process later
        process = subprocess.Popen(command_launch)
        # Sleep for a bit so the process has time to start
        time.sleep(wait)

        command_export = [
            "ros_d2",
            "export",
            str(output_file),
            "--verbose",
        ]
        click.echo(f"Running: {' '.join(command_export)}")
        subprocess.run(command_export)

        # Kill the process
        process.kill()


if __name__ == "__main__":
    cli()



# -*- coding: utf-8 -*-
from pathlib import Path

from ros_d2.helpers.convert import convert
from ros_d2.helpers.get_ros_architecture import get_ros_architecture


def export(output_file: Path, verbose: bool = False):
    ros_architecture = get_ros_architecture()
    d2_graph = convert(ros_architecture, verbose)
    open(output_file, "w+").write(d2_graph)



# -*- coding: utf-8 -*-
from typing import List

from py_d2.D2Connection import D2Connection, Direction
from py_d2.D2Diagram import D2Diagram
from py_d2.D2Shape import D2Shape, D2Text, Shape
from py_d2.D2Style import D2Style

from ros_d2.helpers.get_ros_architecture import RosArchitecture

node_style = D2Style(
    opacity=0.6,
    three_d=True,
)


def concat_strings(strings: List[str]) -> str:
    """Concatenates a list of strings into a single string."""
    return ", ".join(strings)


def convert(ros_architecture: RosArchitecture, verbose: bool = False) -> str:
    """Converts a list of NodeInfo objects to a d2 graph."""
    diagram = D2Diagram()

    for topic in ros_architecture.topics:
        if verbose:
            shape = D2Shape(
                name=topic.name,
                shape=Shape.classs,
                topic=D2Text(text=concat_strings(topic.types), format=""),
            )
        else:
            shape = D2Shape(name=topic.name, shape=Shape.classs)
        diagram.add_shape(shape)

    for node_info in ros_architecture.nodes:
        shape = D2Shape(name=node_info.name, style=node_style)
        diagram.add_shape(shape)
        for sub in node_info.subs:
            diagram.add_connection(
                D2Connection(
                    # label=concat_strings(sub.types),
                    shape_1=node_info.name,
                    shape_2=sub.name,
                    direction=Direction.FROM,
                )
            )
        for pub in node_info.pubs:
            diagram.add_connection(
                D2Connection(
                    # label=concat_strings(pub.types),
                    shape_1=node_info.name,
                    shape_2=pub.name,
                    direction=Direction.TO,
                )
            )
        for client in node_info.clients:
            diagram.add_connection(
                D2Connection(
                    # label=concat_strings(pub.types),
                    shape_1=node_info.name,
                    shape_2=client.name,
                    direction=Direction.TO,
                )
            )
        for service in node_info.services:
            diagram.add_connection(
                D2Connection(
                    # label=concat_strings(pub.types),
                    shape_1=node_info.name,
                    shape_2=service.name,
                    direction=Direction.FROM,
                )
            )

    return str(diagram)



# -*- coding: utf-8 -*-
from dataclasses import dataclass
from typing import List, Tuple

from ros2cli.node.strategy import NodeStrategy
from ros2node.api import get_node_names

TOPICS_TO_IGNORE = [
    "/parameter_events",
    "/rosout",
]

TOPIC_TYPES_TO_IGNORE = [
    "rcl_interfaces/srv/ListParameters",
    "rcl_interfaces/srv/SetParameters",
    "rcl_interfaces/srv/GetParameterTypes",
    "rcl_interfaces/srv/GetParameters",
    "rcl_interfaces/srv/SetParametersAtomically",
    "rcl_interfaces/srv/DescribeParameters",
]


@dataclass
class TopicInfo:
    # Topic name
    name: str
    # Topic types
    types: List[str]


@dataclass
class NodeInfo:
    # Node name
    name: str
    # Node publisher topics
    subs: List[TopicInfo]
    # Node subscriber topics
    pubs: List[TopicInfo]
    # Node services
    services: List[TopicInfo]
    # Node service clients
    clients: List[TopicInfo]


@dataclass
class RosArchitecture:
    nodes: List[NodeInfo]
    topics: List[TopicInfo]


def filter_topics(topics: List[TopicInfo]) -> List[TopicInfo]:
    return [
        t
        for t in topics
        if t.name not in TOPICS_TO_IGNORE and t.types[0] not in TOPIC_TYPES_TO_IGNORE
    ]


def topic_tuple_to_info(topic_tuple: Tuple[str, List[str]]) -> TopicInfo:
    return TopicInfo(name=topic_tuple[0], types=topic_tuple[1])


def get_ros_architecture() -> RosArchitecture:

    # Get all the nodes and their oublisher and subscriber topics
    nodes: List[NodeInfo] = []
    with NodeStrategy([]) as node:
        node_names = get_node_names(node=node)
        for n in node_names:
            node_name = n.full_name
            subs_tuple = node.get_subscriber_names_and_types_by_node(
                n.name, n.namespace
            )
            pubs_tuple = node.get_publisher_names_and_types_by_node(n.name, n.namespace)
            subs = [topic_tuple_to_info(t) for t in subs_tuple]
            pubs = [topic_tuple_to_info(t) for t in pubs_tuple]
            services_tuple = node.get_service_names_and_types_by_node(
                n.name, n.namespace
            )
            services = [topic_tuple_to_info(t) for t in services_tuple]
            clients_tuple = node.get_client_names_and_types_by_node(n.name, n.namespace)
            clients = [topic_tuple_to_info(t) for t in clients_tuple]

            subs = filter_topics(subs)
            pubs = filter_topics(pubs)
            services = filter_topics(services)
            clients = filter_topics(clients)

            node_info = NodeInfo(
                name=node_name, subs=subs, pubs=pubs, services=services, clients=clients
            )
            nodes.append(node_info)

    # Get all the unique topics
    topic_names = set()
    topics: List[TopicInfo] = []
    for node_info in nodes:
        for sub in node_info.subs:
            if sub.name not in topic_names:
                topic_names.add(sub.name)
                topics.append(sub)
        for pub in node_info.pubs:
            if pub.name not in topic_names:
                topic_names.add(pub.name)
                topics.append(pub)
        for srv in node_info.services:
            if srv.name not in topic_names:
                topic_names.add(srv.name)
                topics.append(srv)

    # Order them alphabetically
    topics.sort(key=lambda t: t.name)
    nodes.sort(key=lambda n: n.name)

    return RosArchitecture(
        topics=topics,
        nodes=nodes,
    )
